Capsule = Rock:NewAddon("Capsule", "LibRockConsole-1.0", "LibRockConfig-1.0", "LibRockComm-1.0", "LibRockDB-1.0", "LibRockTimer-1.0", "LibRockEvent-1.0", "LibRockHook-1.0", "FuBarPlugin-2.0")


local self = Capsule
local VERSION = tonumber(("$Revision: 52680 $"):match("%d+"))
self.revision = VERSION
self.date = ("$Date: 2007-10-20 11:45:46 -0400 (Sat, 20 Oct 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
self.svnURL = "$URL: svn://dev.wowace.com/wowace/trunk/Capsule/Capsule.lua $"
local _,_,source = self.svnURL:find("^.+/(.+)/Capsule/")
self.source = source:gsub("^trunk$", "stable")
self.version = "1.0r" .. VERSION .." (".. self.source ..")"

--Fubar
self.hasIcon = "Interface\\Icons\\INV_Misc_EngGizmos_15"
self.hasNoColor  = true
self.cannotDetachTooltip = true
self.blizzardTooltip = true
self.hideWithoutStandby = true
self.title = "Capsule"
self.independentProfile = true

self.chooseChar = nil
self.chooseCustom = nil
self.recipients = {}

--Version
Rock("LibRockComm-1.0"):AddAddonVersionReceptor(function (...)
	Capsule:OnVersionReceive(...)
end)

function self:OnVersionReceive(player, addon, version)
	if (addon == "Capsule") then
		if (version == true) then
			version = "Unknown version"
		elseif (version == false) then
			version = "Not loaded"
		end
		self:Print("Version ("..player.."): "..version)
	end
end

function self:QueryVersion(target)
	if (target:upper() == "TARGET") then target = UnitName("target") end
	if (target ~= nil) then
		target = strtrim(target)
		if (target ~= "") then
			local distribution = target:upper()
			if (distribution ~= "GROUP" and distribution ~= "GUILD") then
				distribution = "WHISPER"
			else
				target = nil
			end
			Rock("LibRockComm-1.0"):QueryAddonVersion("Capsule", distribution, target)
		end
	end
end

self.options = {
	type = 'group',
	name = "Capsule",
	desc = "Saves action buttons and macros into profiles.",
	handler = self,
	icon = "Interface\\Icons\\INV_Misc_EngGizmos_15",
	args = {
		{
			type = 'choice',
			name = 'Allow incoming',
			desc = 'Allow incoming profiles from other users',
			get = function() 
				if (not self.allowIncoming) then
					return "Disabled"
				elseif (self.db.profile.allowIncoming) then
					return "Enabled"
				else
					return "Enabled (this session only)"
				end
			end,
			set = function(val) 
				if (val == "Disabled") then
					self.allowIncoming = false
					self.db.profile.allowIncoming = false
				elseif (val == "Enabled") then
					self.allowIncoming = true
					self.db.profile.allowIncoming = true
				else
					self.allowIncoming = true
					self.db.profile.allowIncoming = false
				end
			end,
			choices = {"Disabled", "Enabled", "Enabled (this session only)"}, 
			order=1,
		},
		{
			type = 'boolean',
			name = 'Protection',
			desc = 'Restores your current setup automatically if it differs on login',
			get = function() return Capsule.db.profile.protection end,
			set = function(val) Capsule.db.profile.protection = val end,
			order=2,
		},
		{
			type = 'execute',
			name = 'List all',
			desc = 'List all custom and character profiles',
			func = "List",
			buttonText = "List",
			order=4,
		},
		{
			type = 'text',
			name = 'Save custom',
			desc = 'Save your current setup as a custom profile',
			get = false,
			set = "SaveCustom",
			usage = "<Profile Name>",
			order=3,
		},
		{
			type = 'execute',
			name = "Show version",
			desc = "Show version info for Capsule.",
			func = "ShowVersion",
			buttonText = "Show version",
			order = 4,
		},
		{
			type = 'text',
			name = "Query version",
			desc = "Query version from other players.",
			get = false,
			set = "QueryVersion",
			validate = function(val)
				if (val ~= strtrim(val)) then
					return false, "Please type a valid name."
				end
			end,
			usage = "<Name|TARGET|GROUP|GUILD>",
			order = 5,
		},
		{
			type = 'group',
			name = 'Custom profiles',
			desc = 'Load and configure custom profiles',
			args = {
				{
					type = 'text',
					name = 'Save',
					desc = 'Save your current setup as a custom profile',
					get = false,
					set = "SaveCustom",
					usage = "<Profile Name>",
					order=1,
				},
				{
					type = 'choice',
					name = 'Load',
					desc = 'Load a custom profile',
					get = false,
					set = "LoadCustom",
					choices = function () return self:ListCustom(true) end,
					order=2,
				},
				{
					type = 'choice',
					name = 'Delete',
					desc = 'Delete a custom profile',
					get = false,
					set = "DeleteCustom",
					choices = function () return self:ListCustom(true) end,
					order=3,
				},
				{
					type = 'execute',
					name = 'Delete all',
					desc = 'Delete all custom profiles',
					func = 'DeleteAllCustom',
					buttonText = "Delete",
					confirmText = "Confirm delete",
					order=4,
				},
				{
					type = "group",
					name = "Send to player",
					desc = "Send a custom profile to another player",
					args = {
						{
							type = 'choice',
							name = 'Choose profile',
							desc = 'Choose a custom profile',
							get = function()
								return self.chooseCustom
							end,
							set = function(val) self.chooseCustom = val end,
							choices = function () return self:ListCustom(true) end,
						},
						{
							type = 'text',
							name = 'Send',
							desc = 'Send a custom profile to another player',
							get = false,
							set = "SendCustom",
							usage = "<Target Player>",
							disabled = function()
								return self.chooseCustom == nil
							end,
							validate = function(val)
								if (val ~= strtrim(val)) then
									return false, "No spaces in the player name"
								elseif (val == "") then
									return false, "Enter a player name"
								end
								return true
							end,
						},
					},
				},
			}
		},
		{
			type = 'group',
			name = 'Character profiles',
			desc = 'Load and configure character profiles',
			args = {
				{
					type = 'choice',
					name = 'Load',
					desc = 'Load a character profile',
					get = false,
					set = "LoadChar",
					choices = function () return self:ListChar(true) end,
					order=1,
				},
				{
					type = 'choice',
					name = 'Delete',
					desc = 'Delete a character profile',
					get = false,
					set = "DeleteChar",
					choices = function () return self:ListChar(true, true) end,
					order=2,
				},
				{
					type = 'execute',
					name = 'Delete all',
					desc = 'Delete all character profiles',
					func = 'DeleteAllChar',
					buttonText = "Delete",
					confirmText = "Confirm delete",
					order=3,
				},
				{
					type = "group",
					name = "Send to player",
					desc = "Send a character profile to another player.",
					order=4,
					args = {
						{
							type = 'choice',
							name = 'Choose profile',
							desc = 'Choose a character profile',
							get = function()
								return self.chooseChar
							end,
							set = function(val) self.chooseChar = val end,
							choices = function () return self:ListChar(true) end,
						},
						{
							type = 'text',
							name = 'Send',
							desc = 'Send a character profile to another player',
							get = false,
							set = "SendChar",
							usage = "<Target Player>",
							disabled = function()
								return self.chooseChar == nil
							end,
							validate = function(val)
								if (val ~= strtrim(val)) then
									return false, "No spaces in the player name"
								elseif (val == "") then
									return false, "Enter a player name"
								end
								return true
							end,
						},
					},
				},
			},
		},
		{
			type = 'execute',
			name = 'Reset',
			desc = 'Reset your setup back from when you login (after Protection is applied)',
			func = "Reset",
			confirmText = "Confirm reset",
			buttonText = "Reset",
		}
	}
}

StaticPopupDialogs["CAPSULE_RECEIVED_SEND"] = {
	text = "%s has sent you a profile. Load it?",
	button1 = YES,
	button2 = NO,
	OnAccept = function(data)
		Capsule:SendCommMessage("WHISPER", data.sender, "ACCEPT_PROFILE")
		Capsule:QueueLoad(data.profile)
	end,
	OnCancel = function(data)
		Capsule:SendCommMessage("WHISPER", data.sender, "REJECT_PROFILE")
	end,
	timeout = 10,
	whileDead = 1,
	hideOnEscape = 1
}

StaticPopupDialogs["CAPSULE_CONFIRM_DELETE_CUSTOM"] = {
	text = "Are you sure you want to delete the custom profile \"%s\"?",
	button1 = YES,
	button2 = NO,
	OnAccept = function(val)
		Capsule:DeleteCustom(val.name, val.returnVal, true)
	end,
	OnShow = function()
		Capsule.configShow = Rock("LibRockConfig-1.0").base:IsShown()
		if (Capsule.configShow) then
			Rock("LibRockConfig-1.0").base:SetFrameStrata("HIGH")
		end
	end,
	onHide = function()
		if (Capsule.configShow) then
			Rock("LibRockConfig-1.0").base:SetFrameStrata("FULLSCREEN_DIALOG")
		end
	end,
	timeout=0,
	exclusive=1,
	whileDead = 1,
	hideOnEscape = 1,
}

StaticPopupDialogs["CAPSULE_CONFIRM_DELETE_CHAR"] = {
	text = "Are you sure you want to delete the character profile \"%s\"?",
	button1 = YES,
	button2 = NO,
	OnAccept = function(val)
		Capsule:DeleteChar(val.name, val.returnVal, true)
	end,
	OnShow = function()
		Capsule.configShow = Rock("LibRockConfig-1.0").base:IsShown()
		if (Capsule.configShow) then
			Rock("LibRockConfig-1.0").base:SetFrameStrata("HIGH")
		end
	end,
	onHide = function()
		if (Capsule.configShow) then
			Rock("LibRockConfig-1.0").base:SetFrameStrata("FULLSCREEN_DIALOG")
		end
	end,
	timeout=0,
	exclusive=1,
	whileDead = 1,
	hideOnEscape = 1,
}

Capsule.allowIncoming = false;
Capsule.realmName = GetRealmName();
Capsule.playerName = UnitName("player");
Capsule.identifier = Capsule.realmName.."."..Capsule.playerName
Capsule.busy = false;
Capsule.numIcons = GetNumMacroIcons();
Capsule.iconTextures = {}
Capsule.origProfile = {}
Capsule.safe = false -- This means that the action buttons may not be trusted!
Capsule.pendingProfiles = {}
Capsule.macroHook = {}
Capsule.macroHookIndex = {}
Capsule.inCombat = false
for i=1,Capsule.numIcons do
	Capsule.iconTextures[GetMacroIconInfo(i)] = i
end

Capsule:SetDatabase("CapsuleDB")
Capsule:SetDatabaseDefaults("profile", {
    char = {},
	custom = {},
	protection = true,
	allowIncoming = false,
} )


function Capsule:OnInitialize()
	-- Called when the addon is loaded
	--self:RegisterComm("boxer", "CUSTOM", self.db.profile.channel, "OnCommReceive")
	--self:SetCommPrefix("boxer")
	--self:RegisterEvent("UNIT_AURA")
	--hooksecurefunc("CastSpellByName", CastSpellByNameHandler)
	--hooksecurefunc("CastSpell", CastSpellHandler)
	self:SetConfigTable(self.options)
	self:SetConfigSlashCommand("/Capsule", "/Cap")
	self:SetCommPrefix("CaP")
end

function Capsule:OnEnable()
	self.allowIncoming = self.db.profile.allowIncoming == true
	self:AddRepeatingTimer(30, "UpdateCurrent")
	self:AddEventListener("LEARNED_SPELL_IN_TAB", "UpdateCurrent")
	self:AddEventListener("CHAT_MSG_SKILL", "UpdateCurrent")
	self:AddEventListener("UPDATE_MACROS", "UpdateCurrent")
	self:AddEventListener("ACTIONBAR_HIDEGRID", "UpdateCurrent")
	self:AddSecureHook("PlaceAction", "UpdateCurrent")
	self:AddSecureHook("PickupAction", "UpdateCurrent")
	self:AddEventListener("TIME_PLAYED_MSG", "Process")
	self:AddEventListener("PLAYER_REGEN_DISABLED", "StartCombat")
	self:AddEventListener("PLAYER_REGEN_ENABLED", "StopCombat")
	self:AddCommListener("capsule", "WHISPER");
	self:DispatchEvent("CAPSULE_LOADED")
end

function Capsule:OnDisable()
	-- Called when the addon is disabled
end

function Capsule:StartCombat()
	self.inCombat = true
end

function Capsule:StopCombat()
	self.inCombat = false
	self:Process()
end


function Capsule:GetActionInfo()
	local output = {}
	for i = 1, 120 do
		if (HasAction(i)) then
			local actionType,actionId,actionSubtype = GetActionInfo(i)
			if (actionType == "spell") then
				local spellName, spellRank = GetSpellName(actionId, actionSubtype)
				tinsert(output, { index=i, type=actionType, id=spellName, spellRank=spellRank, subtype=actionSubtype})
			elseif (actionType == "macro") then
				local name, texture, macrobody, localvar = GetMacroInfo(actionId)
				texture = self.iconTextures[texture]
				if (texture == nil) then texture = 1 end
				if (actionId <= 18) then actionSubtype = "global" else actionSubtype="character" end
				tinsert(output, { index=i, type=actionType, id=name, subtype=actionSubtype, texture=texture,macrobody=macrobody,localvar=localvar})
			elseif (actionType == "item") then
				tinsert(output, { index=i, type=actionType, id=actionId, subtype=actionSubtype, texture })
			end
		end
	end
	return output
end

function Capsule:GetMacroInfo()
	local output = {}
	for i=1, 36 do
		local name, texture, macrobody, localvar = GetMacroInfo(i)
		texture = self.iconTextures[texture]
		if (texture == nil) then texture = 1 end
		if (name ~= nil) then
			local type = "global"
			if (i > 18) then type = "character" end
			table.insert(output, {name=name, texture=texture, macrobody=macrobody, localvar=localvar, type=type})
		end
	end
	return output
end


--Processes all pending profiles to be loaded.
function Capsule:Process()
	--Can only change actions when out of combat
	if (not self.inCombat) then
		if (not self.safe) then
			local char = self.db.profile.char[self.identifier]
			--Apply Protection
			if (self.db.profile.protection and char ~= nil and char.macroInfo ~= nil and char.actionInfo ~= nil) then
				self.safe = true
				self:Load(char, true)
			else
				--Either Protection was not enabled, or this is a new character
				self.safe = true
				self:UpdateCurrent()
			end
			self.origProfile = self.db.profile.char[self.identifier]
			self:DispatchEvent("CAPSULE_SAFE")
		end
		if (self.safe and not self.busy) then
			if (getn(self.pendingProfiles) > 0) then 
				for i,v in ipairs(self.pendingProfiles) do
					self:Load(v)
				end
				self.pendingProfiles = {}
				self:DispatchEvent("CAPSULE_READY")
			end
		end
	end
end

function Capsule:UpdateCurrent()
	if (not(self.busy) and (self.safe) and not self.inCombat) then
		self.busy = true
		for i=1,18 do
			local name, texture, macrobody, localvar = GetMacroInfo(i)
			if name == nil then break end
			for i,v in pairs(self.macroHookIndex) do
				macrobody = self.macroHook[i](macrobody)
			end
			texture = self.iconTextures[texture]
			EditMacro(i, name, texture, macrobody, localvar, 0)
		end
		for i=19,36 do
			local name, texture, macrobody, localvar = GetMacroInfo(i)
			if name == nil then break end
			for i,v in pairs(self.macroHookIndex) do
				macrobody = self.macroHook[i](macrobody)
			end
			texture = self.iconTextures[texture]
			EditMacro(i, name, texture, macrobody, localvar, 1)
		end
		self.busy = false
		local actionInfo = self:GetActionInfo()
		local macroInfo = self:GetMacroInfo()
		self.db.profile.char[self.identifier] = {actionInfo=actionInfo, macroInfo=macroInfo}
		self:DispatchEvent("CAPSULE_UPDATED")
		if (self.pendingProfiles[1] == nil) then
			self:DispatchEvent("CAPSULE_READY")
		end
	end
end

function Capsule:ClearAllActions()
	for i=1,120 do
		PickupAction(i)
		ClearCursor()
	end
end

function Capsule:ClearAllMacros()
	while (true) do
		local name, texture, macrobody, localvar = GetMacroInfo(1)
		if (name ~= nil) then
			DeleteMacro(1)
		else
			break
		end
	end
	while (true) do
		local name, texture, macrobody, localvar = GetMacroInfo(19)
		if (name ~= nil) then
			DeleteMacro(19)
		else
			break
		end
	end
end

function Capsule:RegisterMacroHook(hookname, hookfunc)
	--Search if it exists
	for i,v in pairs(self.macroHookIndex) do
		if (v == hookname) then
			return
		end
	end
	--Not found create new one
	tinsert(self.macroHook, hookfunc)
	tinsert(self.macroHookIndex, hookname)
end

function Capsule:UnregisterMacroHook(hookname)
	--Search if it exists
	for i,v in pairs(self.macroHookIndex) do
		if (v == hookname) then
			tremove(self.macroHookIndex, i)
			tremove(self.macroHook, i)
		end
	end
end

function Capsule:Load(data)
	if (self.busy or self.inCombat) then return end
	self.busy = true
	
	ClearCursor()
	
	--Do Macros first
	self:ClearAllMacros()
	
	local macros = data.macroInfo
	for i, macro in pairs(macros) do
		local isGlobal = 0
		if (macro.type == "character") then isGlobal = 1 end
		CreateMacro(macro.name, macro.texture, macro.macrobody, macro.localVar, isGlobal)
	end
	--for i,macro	in pairs(macros) do
	--	local isGlobal = 0
	--	if (macro.type == "character") then isGlobal = 1 end
	--	if (strict == true) then
	--		while (true) do
	--			local id = Capsule:FindMacro(macro.name, macro.type)
	--			if (id == false) then break end
	--			DeleteMacro(id)
	--		end
	--		CreateMacro(macro.name, macro.texture, macro.macrobody, macro.localVar, isGlobal)
	--	else
	--		if (Capsule:FindMacro(macro.name, macro.type) == false) then
	--			CreateMacro(macro.name, macro.texture, macro.macrobody, macro.localVar, isGlobal)
	--		end
	--	end
	--end
	
	--Do Action Buttons Now
	
	--Make a spell cache first
	local spellCache = {}
	local petCache = {}
	local i = 1
	while true do
		local spellName, spellRank = GetSpellName(i, "spell")
		local petName, petRank = GetSpellName(i, "pet")
		if (not spellName and not petName) then
			break
		end
		if (spellName ~= nil) then
			spellCache[spellName..spellRank] = i
		end
		if (petName ~= nil) then
			petCache[petName..petRank] = i
		end
		i = i + 1
	end
	
	--Make an item cache next
	local bags = {}
	table.insert(bags, 0) --Backpack
	for i=1, NUM_BAG_SLOTS do
		table.insert(bags, i) --Bags
	end
	table.insert(bags, KEYRING_CONTAINER) --Keyring
	table.insert(bags, BANK_CONTAINER) --Bank Window
	for i=NUM_BAG_SLOTS+1, NUM_BAG_SLOTS+NUM_BANKBAGSLOTS do
		table.insert(bags, i) --Bank bags
	end
	local itemCache = {}
	for i,bag in pairs(bags) do
		for j=1, GetContainerNumSlots(bag) do
			local itemLink = GetContainerItemLink(bag, j)
			if (itemLink ~= nil) then
				local s,f = string.find(itemLink, "item:(%d+)")
				s = s + 5
				local itemId = tostring(string.sub(itemLink, s,f))
				if (itemCache[itemId] == nil) then
					itemCache[itemId] = {source="bag", bag=bag, slot=j}
				end
			end
		end
	end
	for i=0,23 do
		local itemLink = GetInventoryItemLink("player", i)
		if (itemLink ~= nil) then
			local s,f = string.find(itemLink, "item:(%d+)")
			s = s + 5
			local itemId = tostring(string.sub(itemLink, s,f))
			if (itemCache[itemId] == nil) then
				itemCache[itemId] = {source="inventory", slot=i}
			end
		end
	end
	
	local used = {}
	for i=1,120 do
		used[i] = false;
	end
	
	local actions = data.actionInfo
	for i,action in pairs(actions) do
		if (action.type == "spell") then 
			local spellId = nil
			if (action.subtype == "spell") then
				spellId = spellCache[action.id..action.spellRank]
			else
				spellId = petCache[action.id..action.spellRank]
			end
			if (spellId == nil) then
				local startPos, endPos = strfind(action.spellRank, "Rank ")
				if (startPos ~= nil) then
					local spellNumRank = tonumber(strsub(action.spellRank, endPos + 1))
					local curId = nil
					for i=1,spellNumRank do
						if (action.subtype == "spell") then
							curId = spellCache[action.id.."Rank " .. i]
						else 
							curId = petCache[action.id.."Rank " .. i]
						end
						if (curId ~= nil) then
							spellId = curId
						else
							break
						end
					end
				end
			end
			if (spellId ~= nil) then
				PickupSpell(spellId, action.subtype)
				PlaceAction(action.index)
				used[action.index] = true
			end
		elseif (action.type == "macro") then
			local macroId = self:FindMacro(action.id, action.subtype)
			if (macroId ~= nil) then
				PickupMacro(macroId)
				PlaceAction(action.index)
				used[action.index] = true
			end
		elseif (action.type == "item") then
			--A little different
			--If the correct item is in the action slot already, but it does not exist
			--in the player's inventory, then we should not clear it.
			local foundItem = itemCache[tostring(action.id)]
			if (foundItem ~= nil) then
				if (foundItem.source == "bag") then
					PickupContainerItem(foundItem.bag, foundItem.slot)
					PlaceAction(action.index)
					used[action.index] = true
				else 
					PickupInventoryItem(foundItem.slot)
					PlaceAction(action.index)
					used[action.index] = true
				end
			else
				local actionType,actionId,actionSubtype = GetActionInfo(action.index)
				if (actionId == action.id) then
					used[action.index] = true
				end
			end
		end
	end
	
	for i=1, 120 do
		if (not used[i]) then
			PickupAction(i)
			ClearCursor()
		end
	end
	self.busy = false
	self:UpdateCurrent();
	
	self:DispatchEvent("CAPSULE_READY")
	--Done
	
end

function Capsule:FindMacro(name, type)
	local low,high = 1,36
	if (type=="character") then
		low = 19
	else
		high=18
	end
	for i=low, high do
		local curName = GetMacroInfo(i)
		if (name == curName) then
			return i
		end
	end
	return false
end

function Capsule:GetMacros()
	
end

function Capsule:List(returnVal)
	if (returnVal == true) then
		local output = {};
		output["custom"] = self:ListCustom()
		output["char"] = self:ListChar()
		return output
	else
		self:ListCustom()
		self:ListChar()
	end
end

function Capsule:ListCustom(returnVal)
	if (returnVal == true) then
		local output = {}
		for index,value in self:SortedPairs(self.db.profile.custom) do
			tinsert(output, index)
		end
		return output
	else
		local msg = "Listing Custom Profiles:"
		for index,value in self:SortedPairs(self.db.profile.custom) do
			msg = msg .. "\n" .. index
		end
		self:Print(msg)
	end
end

function Capsule:ListChar(returnVal, noPlayer)
	if (returnVal == true) then
		local output = {}
		for index,value in self:SortedPairs(self.db.profile.char) do
			if (index ~= self.identifier or noPlayer ~= true) then
				tinsert(output, index)
			end
		end
		return output
	else
		local msg = "Listing Character Profiles:"
		for index,value in self:SortedPairs(self.db.profile.char) do
			msg = msg .. "\n" .. index
		end
		self:Print(msg)
	end
end

function Capsule:SaveCustom(name, returnVal)
	if (name ~= "") then
		self.db.profile.custom[name] = self:CloneTable(self.db.profile.char[self.identifier])
		if (returnVal == true) then
			return true
		else
			self:Print("Saved custom profile: " .. name)
		end
	else
		return false
	end
end

function Capsule:LoadCustom(name, returnVal)
	local target = self.db.profile.custom[name]
	if (target ~= nil) then
		self:QueueLoad(self:CloneTable(target))
		if (returnVal == true) then
			return true
		else
			self:Print("Loading custom profile: " .. name)
		end
	else
		return false
	end
end

function Capsule:LoadChar(name, returnVal)
	local target = self.db.profile.char[name]
	if (target == nil) then
		target = self.db.profile.char[self.realmName.."."..name]
	end
	if (target ~= nil) then
		self:QueueLoad(self:CloneTable(target))
		if (returnVal == true) then
			return true
		else
			self:Print("Loading character profile: " .. name)
		end
	else
		return false
	end
end

function Capsule:LoadCustomStrict(name, returnVal)
	local target = self.db.profile.custom[name]
	if (target ~= nil) then
		self:QueueLoad(self:CloneTable(target), true)
		if (returnVal == true) then
			return true
		else
			self:Print("Loading custom profile (strict): " .. name)
		end
	else
		return false
	end
end

function Capsule:LoadCharStrict(name, returnVal)
	local target = self.db.profile.char[name]
	if (target == nil) then
		target = self.db.profile.char[self.realmName.."."..name]
	end
	if (target ~= nil) then
		self:QueueLoad(self:CloneTable(target), true)
		if (returnVal == true) then
			return true
		else
			self:Print("Loading character profile (strict): " .. name)
		end
	else
		return false
	end
end

function Capsule:DeleteCustom(name, returnVal, confirm)
	if (confirm ~= true) then
		local dialog = StaticPopup_Show("CAPSULE_CONFIRM_DELETE_CUSTOM", name)
		if (dialog) then
			dialog.data = {name=name, returnVal=returnVal}
		end
	else
		local target = self.db.profile.custom[name]
		if (target ~= nil) then
			self.db.profile.custom[name] = nil
			if (returnVal == true) then
				return true
			else
				self:Print("Deleted custom profile: " .. name)
			end
		else
			return false
		end
		Rock("LibRockConfig-1.0"):RefreshConfigMenu(Capsule)
	end
end

function Capsule:DeleteChar(name, returnVal, confirm)
	if (confirm ~= true) then
		local dialog = StaticPopup_Show("CAPSULE_CONFIRM_DELETE_CHAR", name)
		if (dialog) then
			dialog.data = {name=name, returnVal=returnVal}
		end
	else
		local target = self.db.profile.char[name]
		if (target == nil) then
			target = self.db.profile.char[self.realmName.."."..name]
		end
		if (target ~= nil) then
			if (self.chooseChar == name) then
				self.chooseChar = nil
			end
			self.db.profile.char[name] = nil
			if (returnVal == true) then
				return true
			else
				self:Print("Deleted character profile: " .. name)
			end
		else
			return false
		end
		Rock("LibRockConfig-1.0"):RefreshConfigMenu(Capsule)
	end
end

function Capsule:DeleteAllCustom()
	for i,v in pairs(self.db.profile.custom) do
		if (self.chooseCustom == i) then
			self.chooseCustom = nil
		end
		self.db.profile.custom[i] = nil
		self:Print("Deleted custom profile: ".. i)
	end
	Rock("LibRockConfig-1.0"):RefreshConfigMenu(Capsule)
end

function Capsule:DeleteAllChar()
	for i,v in pairs(self.db.profile.char) do
		if (i ~= self.identifier) then
			if (self.chooseChar == i) then
				self.chooseChar = nil
			end
			self.db.profile.char[i] = nil
			self:Print("Deleted character profile: ".. i)
		end
	end
	Rock("LibRockConfig-1.0"):RefreshConfigMenu(Capsule)
end

function Capsule:SendCustom(value, returnVal)
	local target = value
	local profilename = self.chooseCustom
	if (profilename ~= nil) then
		profilename = strtrim(profilename)
		local profile = self.db.profile.custom[profilename]
		if (profile ~= nil) then
			self:Send(target, self:CloneTable(profile), returnVal)
		else
			self:Print("Unable to find profile "..profilename)
		end
	end
end

function Capsule:SendChar(value, returnVal)
	local target = value
	local profilename = self.chooseChar
	if (profilename ~= nil) then
		profilename = strtrim(profilename)
		local profile = self.db.profile.char[profilename]
		if (profile ~= nil) then
			self:Send(target, self:CloneTable(profile), returnVal)
		else
			self:Print("Unable to find profile "..profilename)
		end
	end
end

function Capsule:Reset()
	self:Load(self.origProfile)
	self:Print("Original profile has been loaded")
end

function Capsule:Send(target, profile, returnVal)
	self:Print("Sending profile to " .. target..".")
	if (self:SendCommMessage("WHISPER", target, "SEND_PROFILE", profile)) then
		self.recipients[target] = true
		if (returnVal == true) then
			return true
		end
	else
		if (returnVal == true) then
			return false
		else
			self:Print("Unable to send profile to "..target..".")
		end
	end
end

function Capsule:GetCustom(name)
	return self.db.profile.custom[name]
end

function Capsule:GetChar(name)
	if (name ~= nil) then
		local output = self.db.profile.char[name]
		if (output ~= nil) then
			return output
		else
			output = self.db.profile.char[self.realmName.."."..name]
			return output
		end
	else
		return self.db.profile.char[self.realmName.."."..self.playerName]
	end
end

function Capsule:OnCommReceive(prefix, sender, distribution, type, message)
	if (type == "ACCEPT_PROFILE") then
		if (self.recipients[sender] == true) then
			self:Print(sender .. " has accepted your profile.")
			self.recipients[sender] = nil
		end
	elseif (type == "REJECT_PROFILE") then
		if (self.recipients[sender] == true) then
			self:Print(sender .. " has rejected your profile.")
			self.recipients[sender] = nil
		end
	elseif (type == "BLOCK_PROFILE") then
		if (self.recipients[sender] == true) then
			self:Print(sender .. " has disabled receiving profiles, blocking your profile.")
			self.recipients[sender] = nil
		end
	elseif (type == "SEND_PROFILE") then 
		if (self.allowIncoming) then
			local dialog = StaticPopup_Show("CAPSULE_RECEIVED_PROFILE", sender)
			if (dialog) then
				dialog.data = {sender=sender,profile=message}
			end
			self:DispatchEvent("CAPSULE_RECEIVED_PROFILE", message, dialog)
		else
			self:SendComMMessage("WHISPER", sender, "BLOCK_PROFILE")
		end
	end
end

function Capsule:CloseReceived(profile, dialog)
	StaticPopup_Hide(dialog, profile)
end

function Capsule:AcceptReceived(profile, dialog)
	StaticPopup_Hide(dialog, profile)
	self:QueueLoad(profile)
end

function Capsule:QueueLoad(profile)
	tinsert(self.pendingProfiles, profile)
	self:Process()
end

function Capsule:SortedPairs(t,comparator)
	local sortedKeys = {}
	table.foreach(t, function(k,v) table.insert(sortedKeys,k) end)
	table.sort(sortedKeys,comparator)
	local i = 0
	local function _f(_s,_v)
		i = i + 1
		local k = sortedKeys[i]
		if (k) then
			return k,t[k]
		end
	end
	return _f,nil,nil
end

function Capsule:IsReady()
	return (self.safe or not self.busy or getn(self.pendingProfiles) == 0)
end

--REMEMBER TO TURN IT OFF AFTER USE!
function Capsule:SetBusy(busy)
	if (busy == true or busy == nil) then
		self.busy = true
	else 
		self.busy = false
	end
end

--Shake that thing
function Capsule:GetBusy()
	return self.busy
end

--Clone a table
function self:CloneTable(value, max, counter)
	if (type(value) ~= "table") then
		return false, "Invalid type."
	end
	if (max ~= nil and type(max) == "number" and max > 0) then
		if (counter == nil) then
			counter = 0
		end
		counter = counter + #value
		if (counter > max) then
			return false, "Too many elements."
		end
	else
		max = nil
	end
	
	local output = {}
	for i,v in pairs(value) do
		if (type(v) ~= "table") then
			--No executable bullshit
			if (type(v) == "function" or type(v) == "thread" or type(v) == "userdata") then
				return false, "Invalid data type "..type(v) .. "."
			end
			output[i] = v
		else
			local result,msg =  self:CloneTable(v, max, counter)
			if (result == false) then
				return false, msg
			else
				output[i] = result
			end
		end
	end
	return output
end

function self:ShowVersion()
	message("Capsule "..self.version)
end

function self:OnClick(button)
	if (button == "LeftButton") then
		self:OpenConfigMenu()
	else
		self:OpenMenu()
	end
end

function self:OnTooltipUpdate()
	GameTooltip:AddLine("|cffffff00 Click|r to configure Capsule.\n|cffffff00 Right-Click|r to set FuBar options.")
end